/**********************************************************************\
    Copyright (c) 2009 Corporation for National Research Initiatives;
                        All rights reserved.
     The CNRI open source license for this software is available at 
                  http://hdl.handle.net/1895.26/1013
------------------------------------------------------------------------
Developer: Sean Reilly <sreilly@cnri.reston.va.us>
Developer: Robert Tupelo-Schneck <schneck@cnri.reston.va.us>
Developer: Tommy Tracy <tjt7a@virginia.edu>
\**********************************************************************/

package net.cnri.apps.doserver;

import net.cnri.dobj.*;
import net.cnri.do_api.*;
import net.handle.hdllib.Util; 
import net.handle.util.StringUtils;
import net.cnri.util.*;
import java.io.*;
import java.net.*;
import java.util.*;
import org.verisign.joid.consumer.*;
import org.eclipse.jetty.servlet.*;
import java.io.IOException;
import javax.servlet.http.*;

/** HTTPConnectionHandler objects run in their own thread and are responsible for
 *  managing a single connection to the server.  Each connection may have
 *  numerous operation invocations which is why each connection handler has
 *  its own RequestHandlerPool to keep track of threads for operation invocation
 *  requests.
 */

class HTTPConnectionHandler
extends HttpServlet
{
	public static final boolean DEBUG = true;
	public static final String DEFAULT_MIME_TYPE = "application/octet-stream";
	public static final String DEFAULT_GET_ELEMENT = "content";
	public static final String DEFAULT_ELEMENT_REF_ATT = "content_element";
  public static final String CONTENT_DISPOSITION_HEADER = "Content-Disposition";
	public static final int RC_OK = 200;
	public static final String RC_STR_OK = "OK";
	public static final int RC_NOTFOUND = 404;
	public static final String RC_STR_NOTFOUND = "Not Found";
	private static final int NONCE_RANDOM_SIZE = 20;
	public static final String XML = "application/xml";
	public static final String HTML = "text/html";
  private static final String PLAINTEXT = "text/plain";
  
  public static final String OPENID_ATTRIBUTE = "caller";
  public static final String OPENID_IDENTITY = "openid.identity";
  public static final int BUFFER_SIZE = 1024;
  
	private Main serverMain;
  private DOClient anonymousDOClient;
  private JoidConsumer joid = new JoidConsumer();
  
	// Constructor; set main, construct headers and parameters maps
	HTTPConnectionHandler(Main main) {
		serverMain = main;
    anonymousDOClient = new DOClient(PKAuthentication.getAnonymousAuth());
	}
  
  
  /**
   * This should be called before starting the HTTP server in order to configure the 
   * servlet parameters.
   */
  void setupServletHolder(ServletHolder servletHolder) {
    servletHolder.setInitParameter("storeClassName", "org.verisign.joid.server.MemoryStore");
    servletHolder.setInitParameter("userManagerClassName", "org.verisign.joid.server.MemoryUserManager");
  }
  
	// HTTP Handler
  public void service(HttpServletRequest request, HttpServletResponse response) 
  throws IOException
  {
    // public void handle(String target, Request request, HttpServletRequest _request,
    //                 HttpServletResponse response) throws IOException, ServletException {
    
    String method = request.getMethod();
    if(!method.equals("GET") && !method.equals("POST") && !method.equals("PUT") && !method.equals("DELETE") && 
       !method.equals("HEAD") && !method.equals("OPTIONS")) return;
    
    //request.setHandled(true);
    //parseGetParams(request, request.getRequestURI(), request.getQueryString());
    String identifier = request.getRequestURI();
    
    response.setStatus(HttpServletResponse.SC_OK);
    
    // if there is an OPENID_IDENTITY parameter, verify it and move on
    if(request.getMethod().equals("GET") && request.getParameter(OPENID_IDENTITY) != null) {
			try {
        AuthenticationResult result = joid.authenticate(convertToStringValueMap(request.getParameterMap()));
        String identity = result.getIdentity();
        if(identity!=null) {
          request.getSession(true).setAttribute(OpenIdFilter.OPENID_ATTRIBUTE, identity);
					Cookie cookie = new Cookie(OPENID_IDENTITY, identity);
					//if(cookieDomain != null) cookie.setDomain(cookieDomain);
          //if(cookieMaxAge != null) cookie.setMaxAge(cookieMaxAge.intValue());
          response.addCookie(cookie);
          
          // redirect to get rid of the long url
          response.sendRedirect(result.getResponse().getReturnTo());
          return;
        }
      } catch(Exception e){
        e.printStackTrace();
        System.err.println("auth failed: " + e.getMessage());
      }
    } else {
      // do standard HTTP Basic authentication
      DOAuthentication auth = authentication(request);
      if(auth!=null && !auth.getID().equals(DOConstants.ANONYMOUS_ID)) {
        if(!checkAuthentication(auth)) {
          unauthorized(response);
          return;
        }
      }
    }
    
    if(identifier.startsWith("/auth/")) handleAuthenticate(request, response);
    else if(identifier.startsWith("/do")) handleRest(request,response);
    else if(identifier.startsWith("/invoke")) handleInvoke(request,response);
    else handleWeb(request, response);
  }
  
  
  
  private void handleAuthenticate(HttpServletRequest request, HttpServletResponse response) throws IOException {
    String method = request.getMethod();
    String path = request.getRequestURI();
    String baseURL = request.getScheme()+"://"+request.getServerName()+":"+request.getServerPort();
    
    HttpSession session = request.getSession(true);
    
    if(method.equals("POST")) {
      if(path.equals("/auth/openid")) {
        //String context = request.getContextPath();
        //if(context!=null) baseURL += context;
        // http://www.myopenid.com/server?
        // openid.trust_root=http%3A%2F%2Flocalhost%3A8080%2Fjoid&
        // openid.ns=http%3A%2F%2Fspecs.openid.net%2Fauth%2F2.0&
        // openid.identity=http%3A%2F%2Fseanreilly.com&openid.claimed_id=http%3A%2F%2Fseanreilly.com&
        // openid.mode=checkid_setup&
        // openid.realm=http%3A%2F%2Flocalhost%3A8080%2Fjoid&
        // openid.assoc_handle=%7BHMAC-SHA1%7D%7B4aa015d7%7D%7BP3eXSg%3D%3D%7D&
        // openid.return_to=http%3A%2F%2Flocalhost%3A8080%2Fjoid
        try {
          String id = request.getParameter("openid_url");
          if(id==null || id.trim().length()<=0) {
            response.sendError(HttpServletResponse.SC_BAD_REQUEST, "Missing OpenID URL parameter (openid_url)");
            return;
          }
          String trustRoot = baseURL.toString();
          
          String s = joid.getAuthUrl(id, baseURL, baseURL);
          response.sendRedirect(s);
        } catch (Throwable e) {
          e.printStackTrace(System.err);
        }
      }
    } else if(method.equals("GET")) {
      if(path.equals("/auth/logout") || path.equals("/login/")) {
        session.removeAttribute(OPENID_ATTRIBUTE);
        session.removeAttribute("user");
        session.invalidate();
        response.sendRedirect(request.getContextPath());
      } else if(path.equals("/auth/login") || path.equals("/login/")) {
        printLoginForm(request, response);
      }
    }
  }

  void printLoginForm(HttpServletRequest req, HttpServletResponse response) {
    response.setStatus(HttpServletResponse.SC_OK);
    response.setContentType("text/html");
    // TODO: need to check cookie signature here!
    String loggedInAs = (String)req.getSession(true).getAttribute(OPENID_ATTRIBUTE);
    try {
      PrintWriter w = response.getWriter();
      w.print("<div id=\"openidauth\">");
      
      if(loggedInAs != null) {
        w.print("<span>You are logged in as: "+safeHTML(loggedInAs)+" - <a href=\"/auth/logout\">Logout</a></span>");
      }
      w.print("<form action=\"/auth/openid\" method=\"post\" id=\"openid_form\">");
      w.print("<input type=\"text\" size=\"30\" value=\"\" name=\"openid_url\" id=\"openid_url\"/>");
      w.print("<div> <input type=\"hidden\" name=\"signin\" value=\"true\"/>");
      w.print("<b>Login with your OpenID URL:</b>");
      w.print("<input type=\"submit\" value=\"Login\"/><br/>");
      w.print("<i>For example: <tt>http://someone.bloghost.com</tt></i>");
      w.print("</form> </div>");
      /*
       w.print("<br/><br/>");
       w.print("<img src=\"http://l.yimg.com/us.yimg.com/i/ydn/openid-signin-yellow.png\" alt=\"Sign in with Yahoo\" onclick=\"submitForm('http://www.yahoo.com');\"/>");
       w.print("<br/><br/>");
       w.print("<img src=\"http://buttons.googlesyndication.com/fusion/add.gif\" alt=\"Sign in with Google\" onclick=\"submitForm('https://www.google.com/accounts/o8/id');\"/>");
       */
      w.close();
    } catch (IOException e) {
      System.err.println("Error showing login form: "+e);
    }
  }
  
  
  private Map/*<String, String>*/ convertToStringValueMap(Map/*<String, String[]>*/ parameterMap) {
		Map/*<String,String>*/ ret = new HashMap();
		Set set = parameterMap.entrySet();
		for (Iterator iter = set.iterator(); iter.hasNext();) {
			Map.Entry mapEntry = (Map.Entry) iter.next();
			String key = (String) mapEntry.getKey();
			String[] value = (String[]) mapEntry.getValue();
			ret.put(key, value[0]);
		}
		return ret;
	}
  
  
  void handleInvoke(HttpServletRequest req, HttpServletResponse response) {
    // TODO
  }
  
  /** 
   * Handle a request (or update) for a DO component as a RESTful resource.  The path to 
   * the resource dictates the object ID and element or attribute name.
   
   The paths are similar to the following:
   <pre>
   /do/objectid/
   /do/objectid/el
   /do/objectid/el/elementid
   /do/objectid/el/elementid/att
   /do/objectid/el/elementid/data
   /do/objectid/att
   </pre>
   
   */
	private void handleRest(HttpServletRequest request, HttpServletResponse response) throws IOException {
    String method = request.getMethod();
    boolean isGet = method.equals("GET") || method.equals("HEAD");
    boolean isPut = method.equals("PUT");
    boolean isPost = method.equals("POST");
    boolean isDelete = method.equals("DELETE");
    
    
    String pathStr = request.getRequestURI();
    if(!pathStr.startsWith("/do")) {
      response.setStatus(HttpServletResponse.SC_NOT_FOUND);
      response.getWriter().write("Unrecognized object path");
      return;
    }
    pathStr = pathStr.substring("/do".length());
    if(pathStr.startsWith("/")) {
      pathStr = pathStr.substring(1);
    }
    
    Repository repo = new HTTPRepositoryInterface(request);
    StringTokenizer pathTokens = new StringTokenizer(pathStr, "/", false);
    
    if(!pathTokens.hasMoreTokens()) {
      if(isGet) {
        // Return XML Document Listing all DOs accessible through this DOR; if non, return empty list
        response.setContentType(XML);
        response.setCharacterEncoding("UTF-8");
        Writer w = response.getWriter();
        w.write("<DOList>");
        try {
          for(Iterator allObj = repo.listObjects(); allObj.hasNext(); ) {
            String objID = String.valueOf(allObj.next());
            w.write("<DO id=\"");
            w.write(URLEncoder.encode(objID));
            w.write("\" />\n");
          }
        } catch (Exception e) {
          System.err.println("Error listing objects: "+e);
          return;
        }
        response.getWriter().write("</DOList>\n");
      } else {
        response.setStatus(HttpServletResponse.SC_METHOD_NOT_ALLOWED);
        response.setHeader("Allow","GET, HEAD, POST");
      }
      return;
    }
    
    String objectID = urlDecode(pathTokens.nextToken());
    DigitalObject digObj = null;
    
    if(isPut) {
      /* Creates a new DO with the speciﬁed handle. If created, the DOR will assign
       * assign the speciﬁed handle to the DO and return it in a response with 
       * HTTP 201 Created. 
       */
      if(!repo.verifyDigitalObject(objectID)) {
        digObj = repo.createDigitalObject(objectID);
        response.setStatus(HttpServletResponse.SC_CREATED);
      }
    }
    
    if(digObj==null) {
      digObj = repo.getDigitalObject(objectID);
    }
    
    response.setCharacterEncoding("UTF-8");
    String token = pathTokens.hasMoreTokens() ? pathTokens.nextToken() : null;
    if(token==null || token.length()<=0 || token.equals("att")) {
      String att = pathTokens.hasMoreTokens() ? urlDecode(pathTokens.nextToken()) : null;
      if ((isPut || isPost) && (pathTokens.hasMoreTokens())) {
        String value = streamToString(request.getInputStream());
        boolean attributeExisted = digObj.getAttribute(att, null)!=null;
        // Replaces the plaintext value of this attribute. If created,returns HTTP 201 Created. 
        digObj.setAttribute(att, value);
        
        if(!attributeExisted) {
          response.setStatus(HttpServletResponse.SC_CREATED);
        }
      } else if(!isGet) {
        response.setStatus(HttpServletResponse.SC_METHOD_NOT_ALLOWED);
        response.setHeader("Allow","GET, HEAD, POST");
      }
      
      // Return XML Document listing the name/value pairs of all attributes, data elements, and data element attributes available
      // This listing does not include the actual data stored inside of data elements; if none return 404          
      
      if(att!=null) {
        response.setContentType(PLAINTEXT);
        // we were given a specific attribute, so return it directly
        response.getWriter().write(digObj.getAttribute(att, ""));
      } else {
        response.setContentType(XML);
        Writer w = response.getWriter();
        w.write("<DOR>\n");
        w.write("  <DO id=\"");
        w.write(URLEncoder.encode(objectID));
        w.write("\" >\n");
        HeaderSet atts = digObj.getAttributes();
        for(Iterator it=atts.iterator(); it.hasNext(); ) {
          HeaderItem attItem = (HeaderItem)it.next();
          w.write("    <ATT name=\"");
          w.write(attItem.getName());
          w.write("\" val=\"");
          w.write(attItem.getValue());
          w.write("\" />\n");
        }
        
        String elementIDs[] = digObj.listDataElements();
        for(int i=0; elementIDs!=null && i<elementIDs.length; i++) {
          DataElement element = digObj.getDataElement(elementIDs[i]);
          atts = element.getAttributes();
          w.write("    <EL id=\"");
          w.write(elementIDs[i]);
          w.write("\">\n");
          for(Iterator it=atts.iterator(); it.hasNext(); ) {
            HeaderItem attItem = (HeaderItem)it.next();
            w.write("      <ATT id=\"");
            w.write(attItem.getName());
            w.write("\" val=\"");
            w.write(attItem.getValue());
            w.write("\" />\n");
          }
          w.write("    </EL>\n");
        }
        w.write("  </DO>\n");
        w.write("</DOR>\n");
      }          
    } else if(token.equals("el")) {
      // we are to return element information
      String elementID = pathTokens.hasMoreTokens() ? pathTokens.nextToken() : null;
      if(elementID==null) {
        // no element ID...  provide a list of elements
        if(isGet) {
          String elementIDs[] = digObj.listDataElements();
          Writer w = response.getWriter();
          w.write("<DOR>\n");
          w.write("  <DO id=\"");
          w.write(safeHTML(objectID));
          w.write("\" >\n");
          for(int i=0; elementIDs!=null && i<elementIDs.length; i++) {
            w.write("    <el id=\"");
            w.write(safeHTML(elementIDs[i]));
            w.write(">\n");
            w.write("</el>\n");
          }
        } else {
          response.setStatus(HttpServletResponse.SC_METHOD_NOT_ALLOWED);
          response.setHeader("Allow", "GET, HEAD");
        }
      } else {
        // an element ID was provided, so we're looking for data or attributes
        DataElement element = digObj.getDataElement(elementID);
        String nextToken = pathTokens.hasMoreTokens() ? pathTokens.nextToken() : null;
        Writer w = response.getWriter();
        if(nextToken==null || nextToken.length()<=0 || nextToken.equals("att")) {
          String att = pathTokens.hasMoreTokens() ? urlDecode(pathTokens.nextToken()) : null;
          if ((isPut || isPost) && (pathTokens.hasMoreTokens())) {
            String value = streamToString(request.getInputStream());
            boolean attributeExisted = element.getAttribute(att, null)!=null;
            // Replaces the plaintext value of this attribute. If created,returns HTTP 201 Created. 
            element.setAttribute(att, value);
            
            if(!attributeExisted) {
              response.setStatus(HttpServletResponse.SC_CREATED);
            }
          } else if(!isGet) {
            response.setStatus(HttpServletResponse.SC_METHOD_NOT_ALLOWED);
            response.setHeader("Allow","GET, HEAD, POST");
          }
          
          if(att!=null) {
            // we were given a specific attribute, so return it directly
            response.setContentType(PLAINTEXT);
            response.getWriter().write(element.getAttribute(att, ""));
          } else {
            HeaderSet elAtts = element.getAttributes();
            for(Iterator it=elAtts.iterator(); it.hasNext(); ) {
              HeaderItem attItem = (HeaderItem)it.next();
              w.write("    <att id=\"");
              w.write(attItem.getName());
              w.write("\" val=\"");
              w.write(attItem.getValue());
              w.write("\" />");
            }
          }
        } else if (nextToken.equals("data")) {
          // next part of URI ended with /data
          if(isGet) {
            InputStream in = element.read();
            OutputStream out = response.getOutputStream();
            byte buf[] = new byte[BUFFER_SIZE];
            int r;
            while((r=in.read(buf))>=0) out.write(buf, 0, r);
          } else if(isPut || isPost) {
            InputStream in = request.getInputStream();
            if(in!=null) {
              element.write(in);
            }
            response.setStatus(HttpServletResponse.SC_OK);
          } else {
            response.setStatus(HttpServletResponse.SC_METHOD_NOT_ALLOWED);
            response.setHeader("Allow","GET, HEAD, PUT, POST");
          }
        } else {
          response.setStatus(HttpServletResponse.SC_NOT_FOUND);
          response.getWriter().write("Unrecognized object path");
        }
      }
      
      
    } else {
      response.setStatus(HttpServletResponse.SC_NOT_FOUND);
      response.getWriter().write("Sorry, cannot find URI: "+safeHTML(pathStr));
    }
  }
                                 
  
    
	
	private void handleWeb(HttpServletRequest request, HttpServletResponse response) throws IOException{
    response.setContentType(HTML);
    Repository repo = new HTTPRepositoryInterface(request);
    
    if(request.getRequestURI().equals("/")) {
			// display the default page
      PrintWriter w = response.getWriter();
      
			w.println("<html>");
			w.println("<head><title>"+safeHTML(serverMain.getServerID())+"</title></head>");
			w.println("<body>");
      printLoginForm(request, response);
      
			w.println("<h1>"+objectHref(serverMain.getServerID())+"</h1>");
			w.print("<form action=\"/show\" method=GET>");
			w.print("Object ID: <input type=text name=\"id\">");
			w.println("</form><br/>");
			w.println("or you can <a href=\"/list\">list the objects</a><br/>");
      w.println("You can also search for objects:");
      w.println("<form action=\"/search\" method=POST>");
			w.println("Search terms: <input type=text name=\"q\">");
			w.println("</form><br/>");
      
			w.print("</body></html>");
		} else if(request.getRequestURI().equals("/list")) {
			listObjects(repo, response);
		} else if(request.getRequestURI().equals("/show")) {
      showObject(repo, request, response);
		} else if(request.getRequestURI().equals("/search")) {
      showSearchResults(repo, request, response);
		} else if(request.getRequestURI().equals("/get")) {
			String objID = (String)request.getParameter("id");
      String elementID = (String)request.getParameter("el");
      DigitalObject digObj = null;
      try {
        digObj = objID!=null && objID.trim().length() > 0 ? repo.getDigitalObject(objID) : null;
      } catch (DOException e) {
        System.err.println("Error accessing object: "+e);
      }
      
      if(digObj==null) {
        response.setStatus(HttpServletResponse.SC_NOT_FOUND);
        PrintWriter w = response.getWriter();
        w.println("<html>");
        w.println("<head><title>Object not found: "+safeHTML(objID)+"</title></head>");
        w.println("<body>");
        w.println("<h1>Object not found: "+safeHTML(objID)+"</h1>");
      } else {
        HeaderSet atts = null;
        if(elementID==null) {
          // no specific element was requested... return the one identified as the 
          // primary representation of the object
          elementID = digObj.getAttribute(DEFAULT_ELEMENT_REF_ATT, DEFAULT_GET_ELEMENT);
        }
        
        DataElement element = digObj.getDataElement(elementID);
        String filename = element.getAttribute(DOConstants.FILE_NAME_ATTRIBUTE, elementID);
        response.setContentType(element.getAttribute(DOConstants.MIME_TYPE_ATTRIBUTE, DEFAULT_MIME_TYPE));
        response.setHeader(CONTENT_DISPOSITION_HEADER, 
                           "inline; filename=\""+safeHTML(filename)+"\";");
        
        long size = element.getSize();
        if(size>=0) response.setContentLength((int)size);
        InputStream in = element.read();
        OutputStream out = response.getOutputStream();
        byte buf[] = new byte[BUFFER_SIZE];
        int r;
        while((r=in.read(buf))>=0) out.write(buf, 0, r);
      }
		} else {
			response.setStatus(HttpServletResponse.SC_NOT_FOUND);
			response.getWriter().println("<html>");
			response.getWriter().println("<head><title>Object not found</title></head>");
			response.getWriter().println("<body><p>Object not found</p></body></html>");
		}
	}
  
  /** Invoke the search operation and return the results in human-readable html */
  private void showSearchResults(Repository repo, HttpServletRequest request, HttpServletResponse response) 
  throws IOException
  {
    String queryStr = request.getParameter("q");
    response.setStatus(HttpServletResponse.SC_OK);
    PrintWriter w = response.getWriter();
    if(queryStr==null) {
      w.println("<html>");
      w.println("<head><title>Invalid Search</title></head>");
      w.println("<body>");
      w.println("<h1>Invalid Search: No query string provided</h1>");
      w.println("</body></html>");
    } else {
      w.println("<html>");
      w.println("<head><title>DO Search</title></head>");
      w.println("<body>");
      w.println("<h1>Search Results for "+safeHTML(queryStr)+"</h1>");
      
      HeaderSet params = new HeaderSet();
      params.addHeader("query", queryStr);
      StreamPair io = repo.performOperation(DOConstants.SEARCH_OP_ID, params);
      io.getOutputStream().close();
      InputStream in = io.getInputStream();
      HeaderSet result = new HeaderSet();
      w.write("<ul class=\"search_results\">\n");
      long numResults = 0;
      while(result.readHeaders(in)) {
        w.write("<li> ");
        w.write(objectHref(result.getStringHeader("objectid", "no_object_id!")));
        w.println("<table class=\"do_attributes\">");
        for(Iterator it=result.iterator(); it.hasNext(); ) {
          HeaderItem att = (HeaderItem)it.next();
          w.println("<tr><td class=\"search_result_key\">"+safeHTML(att.getName())+"</td>");
          w.println("<td class=\"search_result_val\">"+safeHTML(att.getValue())+"</td></tr>");
        }
        w.println("</table>\n");
        w.write("</li>");
        numResults++;
      }
      w.write("</ul>\n");
      w.write("Total Results: "+numResults+"<br/>");
      w.write("</body></html>");
    }
  }

  //Called from Web read-only interface
  private void showObject(Repository repo, HttpServletRequest req, HttpServletResponse response) 
  throws IOException
  {
    String objID = req.getParameter("id");
    
    if(objID==null || objID.trim().length()<=0) {
      response.setStatus(HttpServletResponse.SC_NOT_FOUND);
      PrintWriter w = response.getWriter();
      w.println("<html>");
      w.println("<head><title>No Object ID provided!</title></head>");
      w.println("<body>");
      w.println("<h1>No Object ID provided!</h1>");
    } else {
      DigitalObject digObj = repo.getDigitalObject(objID);
      PrintWriter w = response.getWriter();
      w.println("<html>");
      w.println("<head><title>"+safeHTML(objID)+"</title></head>");
      w.println("<body>");
      w.println("<h1>"+safeHTML(objID)+"</h1>");
      try {
        HeaderSet atts = digObj.getAttributes();
        w.println("Object Attributes:<br/>");
        w.println("<table class=\"do_attributes\">");
        for(Iterator it=atts.iterator(); it.hasNext(); ) {
          HeaderItem att = (HeaderItem)it.next();
          w.println("<tr><td class=\"do_attribute_key\">"+safeHTML(att.getName())+"</td>");
          w.println("<td class=\"do_attribute_val\">"+safeHTML(att.getValue())+"</td></tr>");
        }
        w.println("</table>\n");
        w.println("Elements:<br/>");
        w.println("<table class=\"do_elements\">");
        String elementIDs[] = digObj.listDataElements();
        for(int didx=0; elementIDs!=null && didx < elementIDs.length; didx++) {
          DataElement element = digObj.getDataElement(elementIDs[didx]);
          w.println("<tr><td>Element:</td><td>"+elementHref(objID, element.getDataElementID())+"</td></tr>");
          w.println("<tr><td valign=TOP>Attributes:</td><td>");
          atts = element.getAttributes();
          w.println("<table class=\"el_attributes\">");
          for(Iterator it=atts.iterator(); it.hasNext(); ) {
            HeaderItem att = (HeaderItem)it.next();
            w.println("<tr><td class=\"el_attribute_key\">"+safeHTML(att.getName())+"</td>");
            w.println("<td class=\"el_attribute_val\">"+safeHTML(att.getValue())+"</td></tr>");
          }
          w.println("</table>");
          w.println("</td></tr>");
          w.println("<tr><td>&nbsp;</td><td>&nbsp;</td></tr>");
        }
        w.println("</table>");
      } catch (Exception e) {
        w.print("<br/>Error showing object: "+safeHTML(String.valueOf(e)));
      }
      w.print("</body></html>");
    }
  }
  
  //Called from Web read-only interface
  private void listObjects(Repository repo, HttpServletResponse response) 
  throws IOException
  {
    response.reset();
    response.setContentType(HTML);
    response.setStatus(HttpServletResponse.SC_OK);
    
    PrintWriter w = response.getWriter();
    w.println("<html><head><title>List Objects in "+safeHTML(repo.getID())+"</title></head>");
    w.println("<body>");
    w.println("<h1>List Objects</h1>");
    try {
      for(Iterator it=repo.listObjects(); it.hasNext(); ) {
        w.print(objectHref(String.valueOf(it.next())));
        w.println("<br/>");
      }
    } catch (Exception e) {
      w.print("<br/>Error listing objects: "+safeHTML(String.valueOf(e)));
    }
    w.println("</body></html>");
  }
  
  private void unauthorized(HttpServletResponse response) {
    System.err.println("Failed to authenticate");
    response.setHeader("WWW-Authenticate", "Basic realm=\"hdl:" + serverMain.getServerID() + "\"");
    response.setStatus(HttpServletResponse.SC_UNAUTHORIZED);
  }
  
  private boolean checkAuthentication(DOAuthentication auth) {
    try {
      return DOConnection.getResolver().checkAuthentication(auth.toHandleAuth());
    }
    catch (Exception e) {
      return false;
    }
  }
  
  private static DOAuthentication authentication(HttpServletRequest request) {
    DOAuthentication auth = (DOAuthentication) request.getAttribute("DOAuthentication");
    if(auth!=null) return auth;
    String authenticateValue = request.getHeader("Authorization");
    if(authenticateValue != null) {   
      String testStr = "";
      //Check if the request contains authorization header
      int space = authenticateValue.indexOf(' ');
      String parts[] = StringUtils.split(authenticateValue, ' ');
      if(parts.length>=2 && parts[0].equalsIgnoreCase("Basic")) {
        String userpass = Util.decodeString(base64Decode(parts[1]));
        String userpassparts[] = StringUtils.split(userpass, ':');
        if(userpassparts.length>=2) {
          auth = new SecretKeyAuthentication(userpassparts[0],
                                             Util.encodeString(userpassparts[1]));
        }
      }
      if(auth!=null) request.setAttribute("DOAuthentication", auth);
      return auth;
    }
    return null;
  }
  
  
	//Read from stream and return string
	private static String streamToString(InputStream input) throws IOException{
		ByteArrayOutputStream bout = new ByteArrayOutputStream();
		byte buf[] = new byte[128];
		int r;
		while((r=input.read(buf))>=0)
			bout.write(buf, 0, r);
		return Util.decodeString(bout.toByteArray());
	}
	
	@SuppressWarnings("deprecation")
	private static final String urlEncode(String s) {
	    try {
	        return URLEncoder.encode(s,"UTF-8");
	    } catch (UnsupportedEncodingException e) {
        return URLEncoder.encode(s);
	    }
	}
	
  public static final String safeHTML(String str) {
    return Template.cgiEscape(str);
  }
  
  private static final String urlDecode(String s) {
    return StringUtils.decodeURLIgnorePlus(s);
  }
  
  private static final String objectHref(String objID) {
    String encID = urlEncode(objID);
    return "<a class=\"objref\" href=\"/show?id="+urlEncode(objID)+"\" >"+safeHTML(objID)+"</a>";
  }
  
  private static final String elementHref(String objID, String elementID) {
    return "<a class=\"getref\" href=\"/get?id="+urlEncode(objID)+
      (elementID==null ? "" : ("&el="+urlEncode(elementID))) +
      "\" >"+safeHTML(elementID)+"</a>";
  }
  
  private static final byte[] base64Decode(String s) {
         ArrayList<Byte> res = new ArrayList<Byte>(s.length()*3/4);
         int len = s.length();
         int pos = 0;
         int i = 0; 
         int acc = -1;
         while(true) {
             int b = -1;
             while (b<0) {
                 if(i>=len) break;
                 char ch = s.charAt(i);
                 i++;
                 if(ch>='A' && ch<='Z') b = ch - 'A';
                 else if(ch>='a' && ch<='z') b = ch - 'a' + 26;
                 else if(ch>='0' && ch<='9') b = ch - '0' + 52;
                 else if(ch=='+') b = 62;
                 else if(ch=='/') b = 63;
                 else if(ch=='=') break;
             }
             if(b<0) {
                 // if(acc>=0) res.add((byte)acc);
                 break;
             }
             if(pos==0) {
                 acc = b << 2;
                 pos = 1;
             }
             else if(pos==1) {
                 acc |= b >>> 4;
                 res.add((byte)acc);
                 acc = (b & 0xF) << 4;
                 pos = 2;
             }
             else if(pos==2) {
                 acc |= b >>> 2;
                 res.add((byte)acc);
                 acc = (b & 0x3) << 6;
                 pos = 3;
             }
             else {
                 acc |= b;
                 res.add((byte)acc);
                 acc = -1;
                 pos = 0;
             }
         }
         int size = res.size();
         byte[] resArray = new byte[size];
         for(i = 0; i<size; i++) resArray[i] = res.get(i);
         return resArray;
    }
  
  
  private String getAuthenticatedIdentity(HttpServletRequest httpRequest) {
    Cookie cookies[] = httpRequest.getCookies();
    for(int i=0; cookies!=null && i<cookies.length; i++) {
      if(cookies[i].getName().equals(OPENID_IDENTITY)) {
        return cookies[i].getValue();
      }
    }
    return DOConstants.ANONYMOUS_ID;
  }
  
  private class HTTPRepositoryInterface
    extends InternalRepository
  {
    private HttpServletRequest request;
    
    private HTTPRepositoryInterface(HttpServletRequest request) 
    throws DOException
    {
      super(serverMain);
      this.request = request;
    }
    
    
    protected DOServerOperationContext getOpContext(String objectID, String operationID,
                                                    HeaderSet params) {
      return new HTTPOperationContext(request, objectID, operationID, params);
    }
    
  }
                                    
  private class HTTPOperationContext
  extends DOServerOperationContext
                                    
                                    
  {
    private HttpServletRequest httpRequest;
    private HttpSession httpSession;
    
    private String callerID;
    private String operationID;
    private String objectID;
    HeaderSet params;
    private HashMap sessionInfo;
    private StorageProxy storage;
    
    private HTTPOperationContext(HttpServletRequest httpRequest, String objectID, 
                                 String operationID, HeaderSet params) {
      this(httpRequest, getAuthenticatedIdentity(httpRequest), objectID, operationID, params);
    }
    
    private HTTPOperationContext(HttpServletRequest httpRequest, String callerID,
                                 String objectID, String operationID, HeaderSet params) {
      this.httpRequest = httpRequest;
      this.httpSession = httpRequest.getSession(true);
      this.callerID = callerID;
      this.operationID = operationID;
      this.objectID = objectID;
      this.params = params==null ? new HeaderSet() : params;
    }
    
    
    public InetAddress getClientAddress() {
      try {
        return InetAddress.getByName(httpRequest.getRemoteAddr());
      } catch (Exception e) {
        return null;
      }
    }
    
    
    
    /** If the client authenticated with a "forwardable" authentication then this will be
      * used to generate a DOClient using that authentication.  For the time being, we will
      * only return an client with anonymous identity. */
    public DOClient getDOClientWithForwardedAuthentication(String clientID) {
      return anonymousDOClient;
    }
    
    // the caller's authentication is established before the operation is invoked
    public boolean authenticateCaller() { return true; }
    
    public String[] getCredentialIDs() { return null; } // no group credentials through HTTP (yet)
    
    public boolean authenticateCredential(String credentialID) { return false; } // no group credentials through HTTP (yet)
    
    public String getCallerID() { return callerID; }
    
    public String getOperationID() { return operationID; }
    
    public String getTargetObjectID() { return objectID; }
    
    public String getServerID() { return serverMain.getServerID(); }
    
    public HeaderSet getOperationHeaders() { return params; }
    
    public synchronized StorageProxy getStorage() {
      return storage!=null ? storage : (storage = new ConcreteStorageProxy(serverMain.getStorage(),
                                                                           serverMain.getServerID(),
                                                                           objectID, null));
    }
    
    
    public void setConnectionMapping(Object mappingKey, Object mappingData) {
      // NOTE: HTTP mappings use the string value of the mappingKey, so be careful!
      if(mappingData==null) {
        httpSession.removeValue(String.valueOf(mappingKey));
      } else {
        httpSession.putValue(String.valueOf(mappingKey), mappingData);
      }
    }
    
    
    public Object getConnectionMapping(Object mappingKey) {
      // NOTE: HTTP mappings use the string value of the mappingKey, so be careful!
      return httpSession.getValue(String.valueOf(mappingKey));
    }
    
    
    /**
     * Performs the specified operation with the identity of the caller, or as 
     * the container repository if the forwarding operations are configured to use
     * the repository's own identity for forwarded operations.
     * If the serverID is null then the DO client that performs this request
     * will resolve the objectID to find the server.
     */
    public void performOperation(String serverID, String objectID,
                                 String operationID, HeaderSet params,
                                 java.io.InputStream input, java.io.OutputStream output)
    throws DOException 
    {
			HTTPOperationContext derivedCtx = new HTTPOperationContext(httpRequest, callerID, 
                                                                 objectID, operationID,
                                                                 params);
			try {
				serverMain.performOperation(derivedCtx, input, output);
			} catch (Exception e) {
				e.printStackTrace(System.err);
				if(e instanceof DOException) throw (DOException)e;
				throw new DOException(DOException.SERVER_ERROR,
                              "Error invoking derived operation: "+e);
			}
      
    }
    
  }
  
  
}